#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <mpi.h>
#include "ant.h"
#include "ant_tools.h"
#include "mpi_tools.h"

unsigned int xFarm, xFarmLocal;
unsigned int yFarm, yFarmLocal;

void boundary_treatment(void);
void print_ants_by_process(void);
void save_waitall(void);

FILE *file_waitall;

struct cell *current_cellulae, *next_cellulae; 
int time; 


void initial(void)
{
  //------source mpi init
  MPI_Comm_size( MPI_COMM_WORLD, &comm_size);
  MPI_Comm_rank( MPI_COMM_WORLD, &my_rank);
  debug_init();
  
  create_comm_tore_1D();
  create_data_struct_mpi();
  distributing_row();
  
  debug("my_width %d  HEIGHT %d  WIDTH %d ", my_width, HEIGHT, WIDTH);  
  
  debug("allocating my cellulae");
  current_cellulae = (struct cell*)calloc(my_width * HEIGHT, sizeof(struct cell));
  next_cellulae = (struct cell*)calloc(HEIGHT * my_width, sizeof(struct cell));
  
  if (!westest_rank) {
    debug("allocating south cellulae");
    ants_to_west = (struct ant_transfert*) calloc(HEIGHT, sizeof(struct ant_transfert));
    ants_from_west = (struct ant_transfert*) calloc(HEIGHT, sizeof(struct ant_transfert));
    my_west_cellulae_info = (struct cell*)calloc(HEIGHT, sizeof(struct cell));
  }
  if (!eastest_rank) {
    ants_to_east = (struct ant_transfert*) calloc(HEIGHT, sizeof(struct ant_transfert));
    ants_from_east = (struct ant_transfert*) calloc(HEIGHT, sizeof(struct ant_transfert));
    my_east_cellulae_info = (struct cell*)calloc(HEIGHT, sizeof(struct cell));
    debug("allocating north cellulae");
  }
  
  debug("init farm and source location"); 
  upper_distance = sqrt( WIDTH*WIDTH + HEIGHT*HEIGHT ) + 100;
  init_farm_location();
  init_source_location();

  if ( (OUTPUT_RESULT) && (my_rank == 0 ) ){
    global_current_cellulae = (struct cell*)calloc(WIDTH * HEIGHT, sizeof(struct cell));
  }
  if (ANTS_BY_PROCESS) global_ants_by_process = (int*)calloc(ITERATION, sizeof(int)); 
  if (SAVE_WAITALL)  time_waitall = (double*)calloc(ITERATION, sizeof(double));

}

int main(int argc, char *argv[])
{
  int x, y, i, id;
  double timeA, timeB;
  double timeIO, enterWait, exitWait;
  double time_elapsed, max_waitall_by_it;

  MPI_Init(&argc, &argv); 
  initial();
  
  memcpy(current_cellulae, next_cellulae, HEIGHT * my_width * sizeof(struct cell));
  for (id=0; id<my_width*HEIGHT; id++) {
    next_cellulae[id].ant_seeking = 0;
    next_cellulae[id].ant_loaded = 0;
  }
  
  timeIO = 0;
  timeA = MPI_Wtime();
  for(i = 0; i < ITERATION; i++)
    {
      debug(" ");
      debug("------------------------ iteration %d  -----------------", i);
      init_bound_value();
      //treatment of bounds
      debug("---------------boundary treatment");
      boundary_treatment();
      
      debug("---------------treatment of inside domain");
      for(x = 1; x < my_width-1; x++) 
	for(y = 0; y < HEIGHT; y++) 
	  algo(x, y); 
  
      enterWait = MPI_Wtime();
      MPI_Waitall(req, request, statuses);
      exitWait = MPI_Wtime();
      max_waitall_by_it = exitWait - enterWait;
      timeIO += exitWait - enterWait;

      //output
      if (OUTPUT_RESULT) {
	rebuild_solution (global_current_cellulae, current_cellulae, my_width);
	if (my_rank == 0) display_current();
      }
      if (ANTS_BY_PROCESS) global_ants_by_process[i] = ants_by_process; 
      if (SAVE_WAITALL) {
	//MPI_Reduce(&max_waitall_by_it, &time_elapsed, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
	time_waitall[i] = time_elapsed;
      }
      
      memcpy(current_cellulae, next_cellulae, HEIGHT * my_width * sizeof(struct cell));
      for (id=0; id<my_width*HEIGHT; id++) {
	next_cellulae[id].ant_seeking = 0;
	next_cellulae[id].ant_loaded = 0;
      }
      
      //updating bound information coming from mpi communication (ants throught bounds and boundaries informations) 
      update_bound_celllulae();
      time++;
    }
  timeB = MPI_Wtime();
  
  {
    double my_time = timeB - timeA;
    MPI_Reduce(&my_time, &time_elapsed, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    if( my_rank == 0 )  printf("comm_size %d \t time_elapsed %g", comm_size, time_elapsed);
    MPI_Barrier(MPI_COMM_WORLD);
    printf("\n (%d) -> Waitall %g\n", my_rank, timeIO);
    
    if (ANTS_BY_PROCESS) print_ants_by_process();
    if (SAVE_WAITALL) save_waitall();
  }
  debug_close();
    
  MPI_Finalize();

  return 0;
}


/* as you can guess, this fonction is used to deal with what happens
 * with the boundaries of the domains of the processes.
 * It just consists into sending and receiving 
 * the extreme collumns to its neighbours.
 * Just take care of the processes that are located
 * on the eastest and the westest part of the map:
 * they have only one neighbour.
 */

void boundary_treatment(void)
{
  int i, dec, tag;
  int mpierrno;
  
  req = 0;

  // send / receive east and west cellulae informations









  for(i=0; i<HEIGHT; i++) algo(0,i);
  for(i=0; i<HEIGHT; i++) algo(my_width-1, i);

  // send / receive the ants that walked across the boundaries

  



  
  if (!eastest_rank){
    for (i=0; i<HEIGHT; i++) {
      if (ants_from_east[i].ant_loaded != 0) debug("Recv in ants_from_east  %d ant(s) loaded", ants_from_east[i].ant_loaded); 
      if (ants_from_east[i].ant_seeking != 0) debug("Recv in ants_from_east  %d ant(s) seeking", ants_from_east[i].ant_seeking);  
    }
  }
  
  if (!westest_rank){
    for (i=0; i<HEIGHT; i++) {
      if (ants_from_west[i].ant_loaded != 0) debug("Recv in ants_from_west  %d ant(s) loaded", ants_from_west[i].ant_loaded); 
      if (ants_from_west[i].ant_seeking != 0) debug("Recv in ants_from_west  %d ant(s) seeking", ants_from_west[i].ant_seeking);  
    }
  }
}


void print_cellulae_info(char *s, struct cell *cellulae){
  fprintf(debug, "%s:            type %d  ant_loaded = %d   ant_seeking = %d  amount = %d \n ", 
	  s, cellulae->type, cellulae->ant_loaded, cellulae->ant_seeking, cellulae->pheromons_level);  
}   


void print_pheromons_state(struct cell *cellulae, char *s)
{
  int i, j;
  fprintf(debug, "%s :\n ", s);
  for(i=0; i< my_width; i++){
    fprintf(debug, "ligne %d : ", i); 
    for (j=0; j< HEIGHT; j++) fprintf(debug,"  %d", cellulae[xyTOid(i, j)].pheromons_level);
    fprintf(debug, "\n");
  }
  fprintf(debug, "\n");
}

void print_ants_by_process(void){
  int i;
  char filename[512]; 
  FILE *ants_file;
  
  sprintf(filename, "ants_by_process-%d", my_rank); 
  ants_file = fopen(filename, "a");           
  for(i=0; i<ITERATION; i++) fprintf(ants_file, "%d\n", global_ants_by_process[i]);    
  fclose(ants_file);
}

void save_waitall(void){
  int i;
  char filename[512]; 
  FILE *waitall_file;
  
  sprintf(filename, "waitall-%d", my_rank); 
  waitall_file = fopen(filename, "a");           
  for(i=0; i<ITERATION; i++) fprintf(waitall_file, "%2.9f\n", time_waitall[i]);    
  fclose(waitall_file);
}
