#ifndef ant_h_
#define ant_h_

#include "mpi.h"

enum direction { DIR_STAY, DIR_EAST, DIR_NORTH, DIR_WEST, DIR_SOUTH };
enum celltype { PATH, SOURCE, ANTFARM };

struct ant_transfert{
  unsigned int ant_loaded;
  unsigned int ant_seeking;
};

struct cell {
  enum celltype type;
  unsigned int ant_loaded;
  unsigned int ant_seeking;
  union {
    unsigned int source_amount;
    unsigned int path_pheromons_level;
  } u;
};
#define amount           u.source_amount
#define pheromons_level  u.path_pheromons_level

struct coord{
  unsigned int x;
  unsigned int y;
};

//---------------MPI declaration-----------------
int* width_info;
MPI_Datatype MPI_CELLULAE, MPI_ANT_TRANSFERT;
int my_rank, my_west_rank, my_east_rank, comm_size, my_width; 
MPI_Comm comm_tore_1D;
MPI_Status statuses[8];
MPI_Request request[8];
int req;

//data storage
struct cell *my_west_cellulae_info, *my_east_cellulae_info;
struct ant_transfert *ants_to_west, *ants_to_east, *ants_from_west, *ants_from_east;
struct cell *global_current_cellulae;
double upper_distance;
int ants_by_process;
int *global_ants_by_process;
double *time_waitall;
//debug 
char filename[512]; 
FILE *debug;

//Define output and debug
#define ANTS_BY_PROCESS 1
#define OUTPUT_RESULT 1
#define DEBUG 0
#define SAVE_WAITALL 1

// input 
//define domain size
#define WIDTH            60
#define HEIGHT           60
#define ITERATION       1000
//define antfarm and food source  
#define XFARM_GLOBAL      28
#define YFARM_GLOBAL      20
#define ANT_QUANTITY      800
#define XSOURCE_GLOBAL    28
#define YSOURCE_GLOBAL    40
#define SOURCE_QUANTITY 80000

#define modulo(a, b)  ( ((a) == -1) ? 0 : ((a)%(b)) )
#define xyTOid(x, y)   ( (x)*HEIGHT + (y) )
int display_current(void);
extern unsigned int xFarm; 
extern unsigned int yFarm; 
extern struct cell *current_cellulae, *next_cellulae; 
extern struct cell *global_current_cellulae; 

#define east_frontier ( x == my_width-1 )
#define west_frontier ( x == 0 )
#define south_frontier ( y == HEIGHT-1 )
#define north_frontier ( y == 0 )

#define eastest_rank ( my_east_rank == -1 ) 
#define westest_rank ( my_west_rank == -1 ) 

#define EAST(x,y)      x + 1 , y
#define NORTH(x,y)     x     , y - 1
#define WEST(x,y)      x - 1 , y
#define SOUTH(x, y)    x     ,  y + 1
#define EAST_ID(x,y)      xyTOid(x+1, y)
#define WEST_ID(x,y)      xyTOid(x-1, y)
#define NORTH_ID(x,y)     xyTOid(x, y-1)
#define SOUTH_ID(x, y)    xyTOid(x, y+1)

#ifdef DEBUG
#define debug(toto...) do {                                         \
     fprintf(debug, "(%d) %s:%d ", my_rank, __FILE__, __LINE__);    \
     fprintf(debug, toto);                                          \
     fprintf(debug, "\n");                                          \
     fflush(debug);                                                 \
} while(0)

#define debug_init() do {                   \
    if( debug )                             \
      return;                               \
    sprintf(filename, "debug-%d", my_rank); \
    debug = fopen(filename, "a");           \
} while(0)                                          

#define debug_close() do {     \
    fclose(debug);             \
} while(0)

#else
#define debug(toto...) do {} while(0)
#define debug_init()   do {} while(0)
#define debug_close()  do {} while(0)
#endif

#endif


