#include <stdlib.h>
#include <png.h>
#include "ant.h"

#define height (5*HEIGHT)
#define width  (5*WIDTH)

struct rgb {
  unsigned char r,g,b;
};

static float amount_scale;
static float ant_scale;
static float pheromon_scale;

static void initialize_rendering(void)
{
  int x, y, id;
  int max_amount;
  int total_ants = 0;

  max_amount = 0;
  for(x = 0; x < WIDTH; x++)
    for(y = 0; y < HEIGHT; y++)
      {
	id = xyTOid(x, y);
	if( (global_current_cellulae[id].type == SOURCE) &&
	    (global_current_cellulae[id].amount > max_amount) )
	  max_amount = global_current_cellulae[id].amount;
	total_ants += global_current_cellulae[id].ant_loaded + 
	  global_current_cellulae[id].ant_seeking;
      }
  amount_scale = 255.0 / (float)max_amount;
  ant_scale    = 255.0 / (float)total_ants;
  pheromon_scale = 255.0 / (float) (WIDTH); 
}

static void render(png_structp png_ptr)
{
  int x, y;
  int id;
  static struct rgb *r = NULL;

  if( r == NULL )
    r = (struct rgb*)calloc(width, sizeof(struct rgb));

  for(y = 0; y < height; y++)
    {
      for(x = 0; x < width; x++)
	{
	  id = xyTOid(x/5, y/5);
	  switch( global_current_cellulae[id].type )
	    {
	    case PATH:
	      if( global_current_cellulae[id].ant_loaded > 0 )
		{
		  r[x].r = 255 - (unsigned char)(ant_scale * (float)global_current_cellulae[id].ant_loaded);
		  r[x].g = 0;
		  r[x].b = 0;
		}
	      else if (global_current_cellulae[id].ant_seeking > 0 )
		{
		  r[x].r = 255 - (unsigned char)(ant_scale * (float)global_current_cellulae[id].ant_seeking);
		  r[x].g = 255 - (unsigned char)(ant_scale * (float)global_current_cellulae[id].ant_seeking);
		  r[x].b = 0;
		}
	      else
		{
		  r[x].r = 255 - (unsigned char)(pheromon_scale * (float)global_current_cellulae[id].pheromons_level);
		  r[x].g = 255 - (unsigned char)(pheromon_scale * (float)global_current_cellulae[id].pheromons_level);
		  r[x].b = 255 - (unsigned char)(pheromon_scale * (float)global_current_cellulae[id].pheromons_level);
		}
	      break;
	      
	    case SOURCE:
	      r[x].r = 0;
	      r[x].g = 255 - (unsigned char)(amount_scale * (float)global_current_cellulae[id].amount);
	      r[x].b = 0;
	      break;
	      
	    case ANTFARM:
	      r[x].r = 0;
	      r[x].g = 0;
	      r[x].b = 255 - (unsigned char)(ant_scale * (float)(global_current_cellulae[id].ant_seeking+global_current_cellulae[id].ant_loaded));
	      break;
	    }
	}
      png_write_row(png_ptr, ( png_bytep )r ) ;
    }
}

int display_current(void)
{
  static int frame;
  char filename[512];
  FILE *f;
  png_structp png_ptr;
  png_infop info_ptr;

  sprintf(filename, "png/ant%04d.png", frame++);
  f = fopen(filename, "w");
  if( !f )
    {
      perror(filename);
      return -1;
    }

  if(frame == 0)
    initialize_rendering();

  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
				    NULL, NULL);
  if (!png_ptr)
    return -1;

  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
    {
      png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
      return -1;
    } 
  
  if (setjmp(png_jmpbuf(png_ptr)))
    {
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(f);
      printf("Error in PNG\n");
      return -1;
    }

  png_init_io(png_ptr, f);

  png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
  png_set_IHDR(png_ptr, info_ptr, width, height,
	       8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
	       PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info(png_ptr, info_ptr);

  render(png_ptr);

  png_write_end(png_ptr, info_ptr);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  fclose(f);

  return 0;
}
